# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from typing import Any, Dict

from model_compression_toolkit.core.common.graph.base_node import BaseNode

# Edge attributes:
EDGE_SOURCE_INDEX = 'source_index'
EDGE_SINK_INDEX = 'sink_index'


class Edge(object):
    """
    Edge in a directed graph.
    """

    def __init__(self,
                 source_node: BaseNode,
                 sink_node: BaseNode,
                 source_index: int,
                 sink_index: int):
        """
        Construct an edge by two nodes and indices.
        Args:
            source_node: Source node of the edge.
            sink_node: Sink node of the edge.
            source_index: Index of the source node of the edge.
            sink_index: Index of the sink node of the edge.
        """
        self.source_node = source_node
        self.sink_node = sink_node
        self.source_index = source_index
        self.sink_index = sink_index

    def get_attributes(self) -> Dict[str, Any]:
        """
        Get edge's attributes as a dictionary mapping an attribute name to its edge value.
        Returns: Edge's attributes.
        """

        return {EDGE_SOURCE_INDEX: self.source_index,
                EDGE_SINK_INDEX: self.sink_index}

    def __eq__(self, other: Any) -> bool:
        """
        Check if this edge is identical to the edge that was passed.
        Args:
            other: An object to compare to this edge.

        Returns:
            Whether the passed object is identical to this edge.
        """

        if isinstance(other, Edge):
            return other.sink_node == self.sink_node and \
                   other.source_node == self.source_node and \
                   other.source_index == self.source_index and \
                   other.sink_index == self.source_index

        return False

    def __repr__(self) -> str:
        """
        A string representing this edge for display.
        """

        return f'{self.source_node.name}:{self.source_index} -> ' \
               f'{self.sink_node.name}:{self.sink_index}'


def convert_to_edge(edge: Any) -> Edge:
    """
    Get an edge in the graph either as an Edge object (source node, destination node,
    source index and destination index) or as a networkx edge with edge data (triplet
    tuple with nodes and edge data) and return it as an Edge object.

    Args:
        edge: Edge in a graph to convert.

    Returns:
        An Edge object containing source node, destination node,
        source index and destination index.
    """

    if isinstance(edge, tuple) and len(edge) == 3:  # networkx edge representation with edge data
        src_node = edge[0]
        dst_node = edge[1]
        edge_data = edge[2]
        return Edge(src_node,
                    dst_node,
                    edge_data[EDGE_SOURCE_INDEX],
                    edge_data[EDGE_SINK_INDEX])

    elif isinstance(edge, Edge):  # it's already an Edge and no change need to be done
        return edge

    raise Exception('Edges list contains an object that is not a known edge format.')



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Dict, Any, Tuple

from model_compression_toolkit import FrameworkInfo
from model_compression_toolkit.core.common.constants import VIRTUAL_ACTIVATION_WEIGHTS_NODE_PREFIX, \
    VIRTUAL_WEIGHTS_SUFFIX, DEFAULT_CANDIDATE_BITWIDTH, VIRTUAL_ACTIVATION_SUFFIX, ACTIVATION, LINEAR, FLOAT_BITWIDTH
from model_compression_toolkit.core.common.graph.base_node import BaseNode
import numpy as np

from model_compression_toolkit.core.common.quantization.candidate_node_quantization_config import \
    CandidateNodeQuantizationConfig


class VirtualSplitNode(BaseNode):
    """
    A class that represents a node that was split from a kernel node (node with weights).
    """

    def __init__(self, origin_node: BaseNode):
        """
        Init a VirtualSplitNode object.

        Args:
            origin_node: The original node from which the new node was split.
        """

        super().__init__(origin_node.name,
                         origin_node.framework_attr,
                         origin_node.input_shape,
                         origin_node.output_shape,
                         origin_node.weights,
                         origin_node.layer_class,
                         origin_node.reuse,
                         origin_node.reuse_group,
                         origin_node.quantization_attr,
                         origin_node.has_activation)

        self.origin_node = origin_node


class VirtualSplitWeightsNode(VirtualSplitNode):
    """
    A class that represents a node that was split from a kernel node (node with weights) and holds the weights of
    the original node. This node contains the original node's weights and the relevant weights candidate quantization
    config.
    """

    def __init__(self, origin_node: BaseNode):
        """
        Init a VirtualSplitWeightsNode object.

        Args:
            origin_node: The original node from which the new node was split.
        """

        super().__init__(origin_node)

        self.name = origin_node.name + VIRTUAL_WEIGHTS_SUFFIX

        self.candidates_quantization_cfg = origin_node.get_unique_weights_candidates()
        for c in self.candidates_quantization_cfg:
            c.activation_quantization_cfg.enable_activation_quantization = False
            c.activation_quantization_cfg.activation_n_bits = DEFAULT_CANDIDATE_BITWIDTH


class VirtualSplitActivationNode(VirtualSplitNode):
    """
    A class that represents a node that was split from a kernel node (node with weights) and holds the activation
    operation of the original node. This node basically does not apply any operation and only holds the relevant
    activation candidate quantization config.
    """

    def __init__(self, origin_node: BaseNode, activation_class: type):
        """
        Init a VirtualSplitActivationNode object.

        Args:
            origin_node: The original node from which the new node was split.
        """

        super().__init__(origin_node)

        self.name = origin_node.name + VIRTUAL_ACTIVATION_SUFFIX
        self.framework_attr = {ACTIVATION: LINEAR}  # TODO: verify that these values work for Pytorch substitution as well (when implementing in Pytorch)
        self.prior_info = origin_node.prior_info
        self.input_shape = origin_node.output_shape  # the kernel output is the activation input
        self.weights = {}
        self.layer_class = activation_class

        self.candidates_quantization_cfg = origin_node.get_unique_activation_candidates()
        for c in self.candidates_quantization_cfg:
            c.weights_quantization_cfg.enable_weights_quantization = False
            c.weights_quantization_cfg.weights_n_bits = DEFAULT_CANDIDATE_BITWIDTH


class VirtualActivationWeightsNode(BaseNode):
    """
    A node that represents a composition of pair of sequential activation node and weights (kernel) node.
    This structure is used for mixed-precision search with bit-operation KPI.
    The node's candidates are the cartesian product of both nodes' candidates.

    Important: note that not like regular BaseNode or FunctionalNode, in VirtualActivationWeightsNode the activation
    candidates config refer to the quantization config of the activation that precedes the linear operation! instead of
    the output of the linear operation.
    It is ok, since this node is not meant to be used in a graph for creating an actual model, but only a virtual
    representation of the model's graph only for allowing to compute the bit-operations KPI in mixed-precision.
    """

    def __init__(self,
                 act_node: BaseNode,
                 weights_node: BaseNode,
                 name: str,
                 framework_attr: Dict[str, Any],
                 input_shape: Tuple[Any],
                 output_shape: Tuple[Any],
                 weights: Dict[str, np.ndarray],
                 layer_class: type,
                 reuse: bool = False,
                 reuse_group: str = None,
                 quantization_attr: Dict[str, Any] = None,
                 has_activation: bool = True,
                 **kwargs):
        """
        Init a VirtualActivationWeightsNode object.

        Args:
            act_node: The original activation node.
            weights_node: The original weights node.
            name: Node's name
            framework_attr: Framework attributes the layer had which the node holds.
            input_shape: Input tensor shape of the node.
            output_shape: Input tensor shape of the node.
            weights: Dictionary from a variable name to the weights with that name in the layer the node represents.
            layer_class: Class path of the layer this node represents.
            reuse: Whether this node was duplicated and represents a reused layer.
            reuse_group: Name of group of nodes from the same reused layer.
            quantization_attr: Attributes the node holds regarding how it should be quantized.
            has_activation: Whether the node has activations that we might want to quantize.
            **kwargs: Additional arguments that can be passed but are not used (allows to init the object with an
                existing node's __dict__).

        """

        super().__init__(name,
                         framework_attr,
                         input_shape,
                         output_shape,
                         weights,
                         layer_class,
                         reuse,
                         reuse_group,
                         quantization_attr,
                         has_activation)

        self.name = f"{VIRTUAL_ACTIVATION_WEIGHTS_NODE_PREFIX}_{act_node.name}_{weights_node.name}"

        self.original_activation_node = act_node
        self.original_weights_node = weights_node

        v_candidates = []
        for c_a in act_node.candidates_quantization_cfg:
            for c_w in weights_node.candidates_quantization_cfg:
                composed_candidate = CandidateNodeQuantizationConfig(activation_quantization_cfg=c_a.activation_quantization_cfg,
                                                                     weights_quantization_cfg=c_w.weights_quantization_cfg)
                v_candidates.append(composed_candidate)

        # sorting the candidates by weights number of bits first and then by activation number of bits (reversed order)
        v_candidates.sort(key=lambda c: (c.weights_quantization_cfg.weights_n_bits,
                                         c.activation_quantization_cfg.activation_n_bits), reverse=True)

        self.candidates_quantization_cfg = v_candidates

    def get_bops_count(self, fw_impl: Any, fw_info: FrameworkInfo, candidate_idx: int) -> float:
        """
        Computes the composed node's (edge) bit-operation count.

        Args:
            fw_impl: A FrameworkImplementation object with framework specific methods.
            fw_info: A FrameworkInfo object with framework specific information,
            candidate_idx: The index of the node's quantization candidate configuration.

        Returns: The BOPS count of the composed node.

        """
        node_mac = fw_impl.get_node_mac_operations(self.original_weights_node, fw_info)
        candidate = self.candidates_quantization_cfg[candidate_idx]
        weights_bit = candidate.weights_quantization_cfg.weights_n_bits if \
            candidate.weights_quantization_cfg.enable_weights_quantization else FLOAT_BITWIDTH
        activation_bit = candidate.activation_quantization_cfg.activation_n_bits if \
            candidate.activation_quantization_cfg.enable_activation_quantization else FLOAT_BITWIDTH
        node_bops = weights_bit * activation_bit * node_mac
        return node_bops



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import ABC

from . import base_matcher


class BaseWalkMatcher(base_matcher.BaseMatcher):
    """
    Base class for walk matchers.
    """
    pass  # pragma: no cover


class WalkMatcherList(BaseWalkMatcher, ABC):
    """
    Initialize a WalkMatcherList object with a list of nodes to match
    in a graph.
    """

    def __init__(self, matcher_list: list):
        self.matcher_list = matcher_list



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np
import tensorflow as tf
if tf.__version__ < "2.6":
    from tensorflow.keras.layers import LayerNormalization, BatchNormalization
else:
    from keras.layers import LayerNormalization, BatchNormalization

from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.graph.base_graph import Graph, BaseNode, OutTensor
from model_compression_toolkit.core.common.graph.graph_matchers import NodeOperationMatcher
from model_compression_toolkit.core.common.constants import REUSE, REUSE_GROUP
from model_compression_toolkit.core.keras.constants import MOVING_VARIANCE, MOVING_MEAN, BETA, \
    GAMMA, EPSILON, AXIS, CENTER, SCALE


class LayerNormDecomposition(common.BaseSubstitution):
    """
    Adds a BatchNorm node after a LayerNormalization node in the graph that applies the LayerNorm's
    gamma & beta weights in the BatchNorm
    """

    def __init__(self):
        """
        Matches LayerNorm node.
        """
        super().__init__(matcher_instance=NodeOperationMatcher(LayerNormalization))

    @staticmethod
    def _get_weight_by_name(_node, w_str):
        """
        get weight by part of weight name
        :param _node: node to search weights in
        :param w_str: part of name to search
        :return: matching weight
        """
        return [k for k in _node.weights.keys() if w_str in k][0]

    def substitute(self,
                   graph: Graph,
                   ln_node: BaseNode) -> Graph:
        """
        Adds a BatchNorm node after a LayerNormalization node in the graph that applies the LayerNorm's
        gamma & beta weights in the BatchNorm

        Args:
            graph: Graph we apply the substitution on.
            ln_node: LayerNorm node to append with BatchNorm

        Returns:
            Graph after applying the substitution.
        """

        # if both scale and center are False (no gamma & beta) the substitution isn't required
        if ln_node.framework_attr[SCALE] is False and ln_node.framework_attr[CENTER] is False:
            return graph

        bn_node_name = f'{ln_node.name}_bn'
        bn_weights = {}

        # get gamma from LayerNorm, if exists
        if ln_node.framework_attr[SCALE]:
            gamma_name = self._get_weight_by_name(ln_node, GAMMA)
            gamma = ln_node.weights[gamma_name].copy()
        else:
            gamma = np.ones((ln_node.input_shape[-1],))
        bn_weights.update({f'{bn_node_name}/{GAMMA}:': gamma})

        # get beta from LayerNorm, if exists
        if ln_node.framework_attr[CENTER]:
            beta_name = self._get_weight_by_name(ln_node, BETA)
            beta = ln_node.weights[beta_name].copy()
        else:
            beta = np.zeros((ln_node.input_shape[-1],))
        bn_weights.update({f'{bn_node_name}/{BETA}:': beta})

        # Set BatchNorm mean & variance to 0 & 1 respectively (normalization done in LayerNorm)
        bn_weights.update({f'{bn_node_name}/{MOVING_MEAN}:': np.zeros((ln_node.input_shape[-1],))})
        bn_weights.update({f'{bn_node_name}/{MOVING_VARIANCE}:': np.ones((ln_node.input_shape[-1],))})

        epsilon = ln_node.framework_attr[EPSILON]
        axis = ln_node.framework_attr[AXIS]

        _reuse_params = {REUSE: ln_node.reuse, REUSE_GROUP: ln_node.reuse_group}
        new_ln_node = BaseNode(ln_node.name, {AXIS: axis, EPSILON: epsilon, CENTER: False, SCALE: False},
                               ln_node.input_shape, ln_node.output_shape, {}, LayerNormalization,
                               **_reuse_params)
        graph.add_node(new_ln_node)
        bn_node = BaseNode(bn_node_name, {AXIS: axis, EPSILON: 0},
                           ln_node.input_shape, ln_node.output_shape, bn_weights, BatchNormalization,
                           **_reuse_params)
        graph.add_node_with_in_edges(bn_node, [new_ln_node])

        # Connect new nodes
        _in_edge = list(graph.in_edges(ln_node))[0]
        graph.add_edge(_in_edge[0], new_ln_node, **graph.get_edge_data(*_in_edge, 0))
        graph.remove_edge(_in_edge[0], ln_node)
        graph.reconnect_out_edges(current_node=ln_node, new_node=bn_node)

        # Finally, remove the LayerNorm node
        graph.remove_node(ln_node, new_graph_outputs=[OutTensor(bn_node, 0)])

        return graph



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import keras.layers
from keras.engine.input_layer import InputLayer
from tensorflow_model_optimization.python.core.quantization.keras.default_8bit.default_8bit_transforms import \
    InputLayerQuantize
from tensorflow_model_optimization.python.core.quantization.keras.graph_transformations import transforms
from tensorflow_model_optimization.python.core.quantization.keras.quantize_config import QuantizeConfig

from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper

from model_compression_toolkit.core.keras.quantizer.mixed_precision.quantization_config_factory import \
  quantization_config_builder_mixed_precision
from model_compression_toolkit.core.keras.quantizer.mixed_precision.selective_quantize_config import SelectiveQuantizeConfig


class InputLayerWrapperTransform(InputLayerQuantize):
    """
    Allows to configure an input layer with QuantizeWrapper given a QuantizeConfig object to wrap it.
    """

    def __init__(self, input_layer, fw_info, quantize_config: QuantizeConfig, custom_objects):
        super(InputLayerWrapperTransform, self).__init__()

        self.input_layer = input_layer
        self.fw_info = fw_info
        self.name = self.input_layer.name
        self.quantize_config = quantize_config
        self.custom_objects = lambda: custom_objects

    def pattern(self):
        return transforms.LayerPattern('InputLayer', config={'name': self.name})

    def replacement(self, match_layer):
        layer_wrapper = QuantizeWrapper(InputLayer(input_shape=self.input_layer.input_shape),
                                        self.quantize_config)
        quantized_layer_name = f"quant_{self.name}"
        cfg_dict = keras.layers.serialize(layer_wrapper)

        cfg_dict['name'] = quantized_layer_name
        cfg_dict['config']['name'] = quantized_layer_name

        layer_node = transforms.LayerNode(cfg_dict, input_layers=[match_layer])

        return layer_node





# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List, Callable, Tuple

import tensorflow as tf
from model_compression_toolkit.gptq.keras.quantizer.configs.weight_quantizer_gptq_config import WeightQuantizeConfig
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper


class SAM:
    """
    This class implements Sharpness-Aware Minimization for Efficiently Improving Generalization (https://arxiv.org/abs/2010.01412)
    """

    def __init__(self, model2quantized,
                 gradient_step: Callable,
                 optimizer_with_param: List[Tuple[List, List[tf.Tensor]]],
                 rho: float = 0.01,
                 eps: float = 1e-12):
        """
        The init function of Sharpness-Aware Minimization  gradient computation class.
        Args:
            model2quantized: Input quantized module
            gradient_step: A function that returns a list of gradients tensors
            optimizer_with_param: A list of optimizer classes to update with the corresponding parameters.
            rho: A floating point number that set the region of smoothness
            eps: A floating point number for numeric stability
        """
        assert rho >= 0.0, f"Invalid rho, should be non-negative: {rho}"
        self.rho = rho
        self.eps = eps
        self.gradient_step = gradient_step

        self.trainable_variables = [p for _, p in optimizer_with_param]
        self.m_var = [len(p) for p in self.trainable_variables]
        self.n_groups = len(self.trainable_variables)
        self.model2quantized = model2quantized
        self.e_ws = [[] for _ in range(len(optimizer_with_param))]

    def _enable_update_step_param(self):
        """
        This function enables the parameter update (update iteration index and gumbel random variable)
        Returns: None

        """
        for layer in self.model2quantized.layers:
            if isinstance(layer, QuantizeWrapper) and isinstance(
                    layer.quantize_config, WeightQuantizeConfig):
                layer.quantize_config.enable_update()

    def _disable_update_step_param(self):
        """
        This function disables the parameter update (update iteration index and gumbel random variable)
        Returns: None

        """
        for layer in self.model2quantized.layers:
            if isinstance(layer, QuantizeWrapper) and isinstance(
                    layer.quantize_config, WeightQuantizeConfig):
                layer.quantize_config.disable_update()

    def _update_w_location(self, gradients: List[List[tf.Tensor]]):
        """
        This function updates the weights position to the highest point
        Args:
            gradients: A list of gradients tensors

        Returns: None

        """

        for g in range(self.n_groups):
            self.e_ws[g].clear()
            grad_norm = tf.linalg.global_norm(gradients[g])
            ew_multiplier = self.rho / (grad_norm + self.eps)
            for i in range(self.m_var[g]):
                e_w = tf.math.multiply(gradients[g][i], ew_multiplier)
                self.trainable_variables[g][i].assign_add(e_w)
                self.e_ws[g].append(e_w)

    def _restore_w_location(self):
        """
        Restore weights to the original position
        Returns: None

        """
        for g in range(self.n_groups):
            for i in range(self.m_var[g]):
                self.trainable_variables[g][i].assign_add(-self.e_ws[g][i])

    def compute_gradients(self, *arg, **kwargs) -> (tf.Tensor, List[List[tf.Tensor]]):
        """
        This function compute the gradients of SAM optimizer
        Args:
            *arg: args to pass to the gradient step functions
            **kwargs: kwargs to pass to the gradient step functions

        Returns: A tensor of the loss value and a list of gradients tensors

        """
        self._enable_update_step_param()
        loss, grad = self.gradient_step(*arg, **kwargs)
        self._update_w_location(grad)
        self._disable_update_step_param()
        loss, grad = self.gradient_step(*arg, **kwargs)
        self._restore_w_location()
        return loss, grad



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np

import model_compression_toolkit.core.common.quantization.quantization_config as qc
from model_compression_toolkit.core.common.constants import MIN_THRESHOLD, THRESHOLD
from model_compression_toolkit.core.common.quantization.quantization_params_generation.error_functions import \
    get_threshold_selection_tensor_error_function, get_threshold_selection_histogram_error_function, _kl_error_histogram
from model_compression_toolkit.core.common.quantization.quantization_params_generation.qparams_search import \
    qparams_symmetric_selection_tensor_search, \
    qparams_symmetric_selection_histogram_search, kl_qparams_symmetric_selection_histogram_search
from model_compression_toolkit.core.common.quantization.quantizers.quantizers_helpers import \
    get_tensor_max
from model_compression_toolkit.core.common.target_platform import QuantizationMethod


def symmetric_selection_tensor(tensor_data: np.ndarray,
                               p: int,
                               n_bits: int,
                               per_channel: bool = False,
                               channel_axis: int = 1,
                               n_iter: int = 10,
                               min_threshold: float = MIN_THRESHOLD,
                               quant_error_method: qc.QuantizationErrorMethod = qc.QuantizationErrorMethod.MSE) -> dict:
    """
    Compute the optimal threshold based on the provided QuantizationErrorMethod to quantize the tensor.
    Different search is applied, depends on the value of the selected QuantizationErrorMethod.

    Args:
        tensor_data: Tensor content as Numpy array.
        p: p-norm to use for the Lp-norm distance.
        n_bits: Number of bits to quantize the tensor.
        per_channel: Whether the quantization should be per-channel or not.
        channel_axis: Output channel index.
        n_iter: Number of iterations to search for the optimal threshold (not used for this method).
        min_threshold: Minimal threshold to use if threshold is too small (not used for this method).
        quant_error_method: an error function to optimize the parameters' selection accordingly.

    Returns:
        Optimal threshold to quantize the tensor in a symmetric manner.
    """

    unsigned_tensor_data = np.abs(tensor_data)
    tensor_max = get_tensor_max(unsigned_tensor_data, per_channel, channel_axis)

    if quant_error_method == qc.QuantizationErrorMethod.NOCLIPPING:
        threshold = get_init_threshold(min_threshold, tensor_max, per_channel)
    else:
        signed = True  # weights are always signed
        error_function = get_threshold_selection_tensor_error_function(QuantizationMethod.SYMMETRIC, quant_error_method, p, norm=False, n_bits=n_bits, signed=signed)
        threshold = qparams_symmetric_selection_tensor_search(error_function,
                                                              tensor_data,
                                                              tensor_max,
                                                              n_bits,
                                                              per_channel,
                                                              channel_axis,
                                                              min_threshold=min_threshold,
                                                              signed=signed)
    return {THRESHOLD: threshold}


def symmetric_selection_histogram(bins: np.ndarray,
                                  counts: np.ndarray,
                                  p: int,
                                  n_bits: int,
                                  min_value: float,
                                  max_value: float,
                                  constrained: bool = True,
                                  n_iter: int = 20,
                                  min_threshold: float = MIN_THRESHOLD,
                                  quant_error_method: qc.QuantizationErrorMethod = qc.QuantizationErrorMethod.MSE) -> dict:
    """
    Compute the optimal threshold based on the provided QuantizationErrorMethod to quantize a histogram.
    Different search is applied, depends on the value of the selected QuantizationErrorMethod.

    Args:
        bins: Bins values of the histogram.
        counts: Bins counts of the histogram.
        p: p-norm to use for the Lp-norm distance (used only for lp threshold selection).
        n_bits: Number of bits to quantize the tensor.
        min_value: Min value (not used for this method).
        max_value: Max value (not used for this method).
        constrained: Whether the threshold should be constrained or not (not used for this method).
        n_iter: Number of iteration ot search for the threshold (not used for this method).
        min_threshold: Minimal threshold to use if threshold is too small (used only for kl threshold selection).
        quant_error_method: an error function to optimize the parameters' selection accordingly.

    Returns:
        Optimal threshold to quantize the histogram a symmetric manner.
    """
    tensor_max = np.max(np.abs(bins)[1:][counts > 0])
    if quant_error_method == qc.QuantizationErrorMethod.NOCLIPPING:
        threshold = get_init_threshold(min_threshold, tensor_max)
    elif quant_error_method == qc.QuantizationErrorMethod.KL:
        # search for KL error is separated because the error method signature is different from the other error methods.
        threshold = kl_qparams_symmetric_selection_histogram_search(_kl_error_histogram,
                                                                    tensor_max,
                                                                    bins,
                                                                    counts,
                                                                    n_bits,
                                                                    min_threshold=min_threshold)
    else:
        error_function = get_threshold_selection_histogram_error_function(QuantizationMethod.SYMMETRIC, quant_error_method, p)
        threshold = qparams_symmetric_selection_histogram_search(error_function,
                                                                 tensor_max,
                                                                 bins,
                                                                 counts,
                                                                 n_bits,
                                                                 min_threshold=min_threshold)
    return {THRESHOLD: threshold}


def symmetric_no_clipping_selection_min_max(bins: np.ndarray,
                                            counts: np.ndarray,
                                            p: int,
                                            n_bits: int,
                                            min_value: float,
                                            max_value: float,
                                            constrained: bool = False,
                                            n_iter: int = 20,
                                            min_threshold: float = MIN_THRESHOLD,
                                            quant_error_method: qc.QuantizationErrorMethod =
                                            qc.QuantizationErrorMethod.NOCLIPPING) -> dict:
    """
    Gets a threshold between min and max numbers.
    If computed threshold is less than min_threshold, min_threshold is returned.

    Returns:
        A constrained threshold of the min/max values.
    """
    return symmetric_selection_histogram(np.asarray([min_value, max_value]),  # histogram with min-max values only
                                         np.asarray([1]),  # passing dummy counts just to make the dimensions work
                                         p,
                                         n_bits,
                                         min_value,
                                         max_value,
                                         constrained,
                                         n_iter,
                                         min_threshold=min_threshold,
                                         quant_error_method=qc.QuantizationErrorMethod.NOCLIPPING)


def get_init_threshold(min_threshold: float, tensor_max: np.ndarray, per_channel: bool = False) -> np.ndarray:
    """
    Gets an initial value for the threshold optimization process.
    If per_channel then returns a vector with initial value for each channel.

    Args:
        min_threshold: Minimal threshold to use if threshold is too small (not used for this method).
        tensor_max: Max value of a tensor.
        per_channel: Whether the quantization should be per-channel or not.

    Returns:
        Threshold value if max value in tensor is larger than min_threshold.
    """
    if per_channel:
        init_t = tensor_max
        init_t[tensor_max < min_threshold] = min_threshold
        return init_t
    return max(min_threshold, tensor_max)
